!*************************************************************
! Module sourcedetection_mod
! Created 15th April 2002- C.Brindle
! Contains a subroutine for detecting "point" sources and one 
! for detecting any other kind of source
!****************************************************************
module centroid_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
!    use minim_mod
    

real(single), pointer, private :: imagePtr(:, :) => null()
real(single), private :: cRadius
integer(int16), private :: cil, ciu, cjl, cju
contains
!**************************************************************************************
!
!*************************************************************************************
subroutine Centroid(source, image, bin)
    type(SourceParameters), intent(inout) :: source
    real(single), pointer :: image(:, :)
    integer, intent(in) :: bin
    integer(int16) :: nx, ny
    real(double), dimension(2) :: x, step
    real(double) :: stopcr, simp, var(2), functionValue 
    integer(int8) :: dx
    integer(int16) :: i0, j0    
    integer :: nop, maxfn, nloop, iquad, ifault


    if (.not. associated(image)) return
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)

    cRadius = 8 / bin
    dx = cRadius + 2 
    cil = max(0_int16, i0 - dx)
    ciu = min(nx, i0 + dx)
    cjl = max(0_int16, j0 - dx)
    cju = min(ny, j0 + dx)

   

    imagePtr => image

  
    step = 0.1_double
    nop = 2
    maxfn = 1000
    stopcr = 0.1_double
    nloop = 2 * nop
    iquad = 1
    x(1) = source%xpos - 1.0
    x(2) = source%ypos - 1.0
    simp = 0.0_double

!    call minim(x, step, nop, functionValue, maxfn, stopcr, nloop, iquad,  &
!                 simp, var, centroidFunction, ifault)


    x = x + 1.0_double

    write(*, *) "Function value=", functionValue, " ifault=", ifault
    write(*, *) "Initial x coord=", source%xpos, " centroided=", x(1)
    write(*, *) "Initial y coord=", source%ypos, " centroided=", x(2)

    source%xpos = x(1)
    source%ypos = x(2)

    nullify(imagePtr)


end subroutine Centroid

!**************************************************************************************
!
!*************************************************************************************
Function CentroidFunction(x) result(counts)
    real(double), intent(inout) :: x(:)   
    real(double) :: counts
    integer(int16) :: i, j, n
    real(double) :: dx, dy, radius


    counts = 0.0_double
    n = 0
    do j = cjl, cju
        dy = x(2) - dble(j)
        if (abs(dy) .gt. cRadius) cycle
        do i = cil, ciu
            dx = x(1) - dble(i)
            radius = sqrt(dx * dx + dy * dy)
            if (radius .le. cRadius) then
                counts = counts + dble(imagePtr(i, j)) * (cRadius - radius) / cRadius
                n = n + 1 
            end if

        end do
    end do

    if (n .gt. 0) then
        counts = -counts / dble(n)
    end if
end Function CentroidFunction

end module centroid_mod


