!*************************************************************
! Module omsourcemap_mod
!
!
!****************************************************************
module omsourcemap_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmConstants_mod      ! For binx, biny, etc
    use omsmoothimage_mod           ! For polynomial smoothing
    use omnumerics_mod 
    use leastsquares_mod
    use ominterpolate_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod


interface SmoothLevelImage
    module procedure SmoothLevelImage1,  SmoothLevelImage2
end interface SmoothLevelImage


interface DilateImage
    module procedure  DilateLevelImage1, dilateLevelimage2, &
                      DilateImage1, dilateimage2

end interface DilateImage



interface RefineSourceMap

    module procedure refinesourcemap1, refinesourcemap2, refineSourceMap3

end interface RefineSourceMap

interface RefineExtendedSourceMap

    module procedure refineExtendedsourcemap1, refineExtendedsourcemap2

end interface RefineExtendedSourceMap

contains
!************************************************
!
!***********************************************
subroutine DilateLevelImage1(image, number)
    integer(int16)  :: image(:, :)
    integer(int16), intent(in) :: number
    integer(int16), pointer :: tempImage(:, :) => null()

    integer(int16) :: nx, ny, i, j, il, jl
    integer(int16) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            maxvalue = image(i, j)
            if (i .gt. il) then ! 
                if (image(i - 1, j) .eq. number) maxValue = number ! Adjacent left pixel
             
                if (j .gt. jl) then
                    if (image(i - 1, j - 1) .eq. number) maxvalue = number! Adjacent bottom-left pixel
                end if        
                if (j .lt. ny) then
                    if (image(i - 1, j + 1) .eq. number) maxvalue = number! Adjacent top-left pixel
                end if        

            end if

            if (j .gt. jl) then
            
                if (image(i, j - 1) .eq. number) maxvalue = number! Adjacent bottom pixel
            end if        
            if (j .lt. ny) then
                if (image(i, j + 1) .eq. number) maxvalue = number! Adjacent top pixel
            end if        

           if (i .lt. nx) then ! 
                if (image(i + 1, j) .eq. number) maxvalue = number! Adjacent right pixel
                if (j .gt. jl) then
                    if (image(i + 1, j - 1) .eq. number) maxvalue = number! Adjacent bottom-right pixel
                end if        
                if (j .lt. ny) then
                    if (image(i + 1, j + 1) .eq. number) maxvalue = number! Adjacent top-right pixel
                end if        

            end if
            
            tempImage(i, j) = maxValue
        

        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateLevelImage1
!************************************************
!
!***********************************************
subroutine DilateLevelImage2(image, il1, iu1, jl1, ju1, dx, number)
    integer(int16), pointer  :: image(:, :)
    integer(int16), intent(in) :: il1, iu1, jl1, ju1
    integer(int16), intent(in) :: dx, number

    integer(int16), pointer :: tempImage(:, :) => null()

    integer(int16) :: nx, ny, i, j, k, l
    integer(int16) :: maxValue, il, iu, jl, ju

    nx = ubound(image, 1)
    ny = ubound(image, 2)
 
    il = max(0_int16, il1 - dx)
    iu = min(nx, iu1 + dx)
    jl = max(0_int16, jl1 - dx)
    ju = min(ny, ju1 + dx)

    tempImage => Allocate2DInt16(int(il, int32), int(iu, int32), int(jl, int32), int(ju, int32), "tempImage")


    tempImage = image(il : iu, jl : ju)


    do j = jl, ju
        do i = il, iu
            maxvalue = image(i, j)

            kloop : do k = max(0_int16, i - dx), min(nx, i + dx)

                do l = max(0_int16, j - dx), min(ny, j + dx)

                    if (image(k, l) .eq. number) then


                        tempimage(i, j) = number

                        exit kLoop
                    end if
                end do 
            end do kloop

        

        end do
    end do


    image(il : iu, jl : ju) = tempImage


    deallocate(tempImage)
    nullify(tempImage)


end subroutine DilateLevelImage2

!************************************************
!
!***********************************************
subroutine ErodeLevelImage(image, number)
    integer(int16)  :: image(:, :)
    integer(int16), intent(in) :: number
 

 
    integer(int16), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, il, jl
    integer(int16) :: minValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            if (image(i, j) .ne. number) cycle
      
            minValue = image(i, j)
            if (i .gt. il) then ! 
                
                if (j .gt. jl) then
                    if (image(i - 1, j - 1) .eq. 0) minValue = 0! Adjacent bottom-left pixel
                end if        
                if (j .lt. ny) then
                    if (image(i - 1, j + 1) .eq. 0) minValue = 0! Adjacent top-left pixel
                end if        

            end if

            if (j .gt. jl) then
            
                if (image(i, j - 1) .eq. 0) minValue = 0 ! Adjacent bottom pixel
            end if        
            if (j .lt. ny) then
                if (image(i, j + 1) .eq. 0) minValue = 0 ! Adjacent top pixel
            end if        

           if (i .lt. nx) then ! 
                if (image(i + 1, j) .eq. 0) minValue = 0! Adjacent right pixel
                if (j .gt. jl) then
                    if(image(i + 1, j - 1) .eq. 0) minValue = 0! Adjacent bottom-right pixel
                end if        
                if (j .lt. ny) then
                    if (image(i + 1, j + 1) .eq. 0) minValue = 0! Adjacent top-right pixel
                end if        

            end if
             
            tempImage(i, j) = minValue
        

        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine ErodeLevelImage


!***********************************************
!
!***********************************************
subroutine DilateImage1(image)
    real(single), pointer  :: image(:, :)
  
  
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    real(single) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DSingle(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            maxValue = image(i, j)
            do k = max(il, i - 1), min(nx, i + 1)
                do l = max(jl, j - 1), min(ny, j + 1)
                    maxValue = max(maxValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = maxValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateImage1


!***********************************************
!
!***********************************************
subroutine DilateImage2(image, dx)
    integer(int16), pointer  :: image(:, :)
    integer(int8), intent(in) :: dx
 

 
    integer(int16), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    integer(int16) :: maxValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DInt16(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            maxValue = image(i, j)
            do k = max(il, i - dx), min(nx, i + dx)
                do l = max(jl, j - dx), min(ny, j + dx)
                    maxValue = max(maxValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = maxValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine DilateImage2

!***********************************************
!
!***********************************************
subroutine ErodeImage(image)
    real(single), pointer  :: image(:, :)
  
 
    real(single), pointer :: tempImage(:, :) => null()

    integer :: nx, ny, i, j, k, l, il, jl
    real(single) :: minValue

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    il = lbound(image, 1)
    jl = lbound(image, 2)


    tempImage => Allocate2DSingle(il, nx, jl, ny, "tempImage")

    tempImage = image

    do j = jl, ny
        do i = il, nx
            minValue = image(i, j)
            do k = max(il, i - 1), min(nx, i + 1)
                do l = max(jl, j - 1), min(ny, j + 1)
                    minValue = min(minValue, image(k, l))
                end do
             end do      
        
        
            tempImage(i, j) = minValue
        end do
    end do
    
    image = tempImage
    deallocate(tempImage)
    nullify(tempImage)

end subroutine ErodeImage
!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage1(image, number, dx, dy)
    
  
    integer(int16), pointer :: image(:, :)
    integer(int16), intent(in) :: number    
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny
     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
    tempImage = image
   
    do j = 0, ny
        
              
        do i = 0, nx
                     
            if (image(i, j) .eq. number .or. image(i, j) .ne. 0) cycle            
           
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                    if (k .eq. i .and. l .eq. j) cycle
                    if (tempImage(k, l) .eq. number) then
                        image(i, j) = number
                        exit kLoop
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

 
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage1



!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage2(image, dx, dy)
    integer(int16), pointer :: image(:, :)
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
   
   
    tempImage = image
    do j = 0, ny
                     
        do i = 0, nx
                     
            if (image(i, j) .eq. int(0, int16)) cycle            
                
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                
                    if (image(k, l) .eq. int(0, int16)) then
                        tempImage(k, l) = image(i, j)
                      
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

    image = tempImage
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage2

!**********************************************************
!
!**********************************************************
subroutine SmoothLevelImage3(image, dx, dy)
    integer(int16), pointer :: image(:, :)
    integer, intent(in) :: dx, dy
    integer i, j, k, l, nx, ny     
    integer(int16), pointer :: tempImage(:, :)
    
  
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    allocate(tempImage(0 : nx, 0 : ny))
   
   
    tempImage = image
    do j = 0, ny
                     
        do i = 0, nx
                     
            if (image(i, j) .ge. int(0, int16)) cycle            
                
            kLoop : do k = max(0, i - dx), min(nx, i + dx)

                do l = max(0, j - dy), min(ny, j + dy)
                
                    if (image(k, l) .eq. int(0, int16)) then
  
                        tempImage(k, l) = 1
                     
                    end if
                end do

            end do  kLoop
                
            
        end do
        
    end do

    image = tempImage
    deallocate(tempImage)

    nullify(tempImage)
     

end subroutine SmoothLevelImage3





!**********************************************************************
! RefineSourceMap1
!
!
!**********************************************************************
subroutine RefineSourceMap1(nSources, sourceList, image, backgroundImage, &
             backgroundSigmaImage, levelImage, nSigma1)
    integer, intent(in) :: nSources
  
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list

    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    real(single), intent(in), optional :: nsigma1

   
    integer(int16) :: i

    real(single), parameter :: nstd = 3.0
    real(single) :: nSigma

    type(sourceParameters), pointer :: source(:) => null()
 

    if (nSources .eq. 0) return

    source => SourceDataArray(sourceList, nSources)
    if (.not. associated(source)) return



    if (present(nsigma1)) then
        nSigma = nSigma1
    else
        nSigma = 2.0
    end if
    call RefineSourceMap2(nSources, source, image, backgroundImage, &
                          backgroundSigmaImage, levelImage, nSigma)

    call deleteSourceList( sourceList )
   

    do i = 1, nSources
        call addSourceToList( sourceList, source(i) )   

    end do
    deallocate(source)
    nullify(source)

   
    

end subroutine refinesourcemap1



!**********************************************************************
! RefineSourceMap2
!
!
!**********************************************************************
subroutine RefineSourceMap2(nSources, source, image, backgroundImage,   &
                            backgroundSigmaImage, levelImage, exposure, &
                            nSigma1)
    integer, intent(in) :: nSources
  
    type(sourceParameters), pointer :: source(:) 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    real(single), intent(in) :: exposure
    real(single), intent(in), optional :: nsigma1

    integer(int16) :: i
    integer(int16) :: nx, ny
    integer(int16) :: dx 


    real(single), parameter :: nstd = 3.0
    real(single) :: nsigma    

    if (nSources .eq. 0) return
    if (.not. associated(source)) return

    nx = ubound(image, 1)
    ny = ubound(image, 2)


  
    write(errstr,'(a)') "Refining source map"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 100 / binx

 
    if (present(nsigma1)) then
        nSigma = nSigma1
    else
        nSigma = 2.0
    end if
 

    do i = 1, nSources

        call RefineSourceMap3(source(i), image, backgroundImage, &
                              backgroundSigmaImage, levelImage,  &
                              nSigma, exposure)
 
    end do

  
end subroutine RefineSourceMap2

!**********************************************************************
! RefineSourceMap3 - For a single source
!
!
!**********************************************************************
subroutine RefineSourceMap3(source, image, backgroundImage, backgroundSigmaImage, &
             levelImage, nsigma, exposure, maskImage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    real(single), intent(in) :: nsigma, exposure
    logical(boolean), pointer, dimension(:, :), optional :: maskImage
   
    integer(int16) :: j
    integer(int16) :: nx, ny
    integer(int16) :: il, iu, jl, ju, dx 
    integer(int16) :: k, l, m, n, dl

    real(single), parameter :: nstd = 3.0
    integer(int32) :: maxImage
    real(single) :: minHeight    

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx = 6 / binx ! was 100


    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)
  
    minHeight = maxval(image(source%imin : source%imax, source%jmin : source%jmax)) 
    maxImage = int(minheight * 1000.0 / exposure)
    minHeight = 0.25 * minheight
!    minHeight = max(1500.0, minheight)
 
    minHeight = 0.1 * minheight

    minHeight = max( 100.0, minheight)
 
!*************PUT IN FOR TEST BFORE CHRISTMAS
minheight =  1.5 * maxval(backgroundimage(source%imin : source%imax, source%jmin : source%jmax))
 


    select case (maxImage)
        case(10000 : )
            dl = 12 / binx
        case (5000 : 9999)
            dl = 10 / binx
        case (2500 : 4999)
            dl = 6 / binx
        case default
            dl = 2 / binx
    end select    

 
    do j = source%imin, source%imax
        do k = source%jmin, source%jmax
            if (levelImage(j, k) .eq. source%number) levelImage(j, k) = 0
        end do
    end do
                           
    source%imin = nx
    source%imax = 0
    source%jmin = ny
    source%jmax = 0


    j = nint(source%xpos - 0.5)
    j = min(j, nx)
    k = nint(source%ypos - 0.5)
    k = min(k, ny)
        
    levelImage(j, k) = source%number

    dx = 100 / binx ! was 100
    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)


    do l = 1, dx
        il = max( int(0, int16), j - l)
        iu = min( nx, j + l)
        jl = max( int(0, int16), k - l)
        ju = min( ny, k + l)


        do n = jl, ju
            do m = il, iu

                if (l .le. dl) then

                    if (l .gt. 4 / binx) then
                        if(image(m, n) .ge. minheight) then
                            cycle
                        end if
                    endif
                    levelImage(m, n) = source%number
                    source%imin = min( source%imin, m)
                    source%imax = max( source%imax, m)
                    source%jmin = min( source%jmin, n)
                    source%jmax = max( source%jmax, n)
                    if (present(maskImage)) maskImage(m, n) = .true.    

                    cycle
                                            
                end if             


                if (levelImage(m, n) .ne. 0) cycle

          
                if (setLevelImagePixel(source, image, backgroundimage, &
                     backgroundSigmaImage, levelimage, m, n, nsigma)) then

                    source%imin = min( source%imin, m)
                    source%imax = max( source%imax, m)
                    source%jmin = min( source%jmin, n)
                    source%jmax = max( source%jmax, n)
                    levelImage(m, n) = source%number
                    if (present(maskImage)) maskImage(m, n) = .true.    

                end if
  
            end do
            
        end do

    end do

  
end subroutine RefineSourceMap3

!**********************************************************************
! RefineSourceMap2
!
!
!**********************************************************************
subroutine RefineExtendedSourceMap1(nSources, source, image, &
                            backgroundImage, backgroundSigmaImage, levelImage, nSigma1)
    integer(int16), intent(in) :: nSources
  
    type(sourceParameters), pointer :: source(:) 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
   
    real(single), intent(in), optional :: nsigma1

    integer(int16) :: i
    integer(int16) :: nx, ny
    integer(int16) :: dx 

    real(single), parameter :: nstd = 3.0

    real(single) :: nsigma    
return
    if (nSources .eq. 0) return
    if (.not. associated(source)) return

    nx = ubound(image, 1)
    ny = ubound(image, 2)

  
    write(errstr,'(a)') "Refining source map"
    call message(AppMsg, Verbose)
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    dx = 100 / binx



 
    if (present(nsigma1)) then
        nSigma = nSigma1
    else
        nSigma = 2.0
    end if
 

    do i = 1, nSources


        call RefineExtendedSourceMap2(source(i), image, backgroundImage, backgroundSigmaImage, levelImage, nSigma)
 
    end do


  
end subroutine RefineExtendedSourceMap1

!**********************************************************************
! RefineExtendedSourceMap - For extended sources
!
!
!**********************************************************************
subroutine RefineExtendedSourceMap2(source, image, backgroundImage, &
             backgroundSigmaImage, levelImage, nsigma)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: image, backgroundImage, backgroundSigmaImage
    integer(int16), pointer, dimension(:, :) :: levelImage
    real(single), intent(in) :: nsigma
    
   
    integer(int16) :: i, j, nx, ny
    integer(int16) :: il, iu, jl, ju, dx 
    integer(int16) :: k, i0, j0
    
    real(single), parameter :: nstd = 3.0
    real(single) :: fMax
    

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx = 10 / binx

    ! Find the location of the maximum pixel value within the source region

    fMax = 0.0
    do j = source%jmin, source%jmax
        do i = source%imin, source%imax
            if (levelImage(i, j) .ne. source%number) cycle
            if (image(i, j) .gt. fmax) then
                fMax = image(i, j)
                i0 = i
                j0 = j
             end if
         end do   
    end do


    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jMin - dx, int(0, int16))
    ju = min(source%jMax + dx, ny)
   
     

    dx = max( max(i0 - il, iu - i0), max(j0 - jl, ju - j0))
    dx = dx + 2

    do i = 1, dx
        il = max(int(0, int16), i0 - i)
        iu = min(nx, i0 + i)
        jl = max(int(0, int16), j0 - i)
        ju = min(ny, j0 + i)

        do j = jl, ju
            do k = il, iu
             
                if (levelImage(k, j) .eq. source%number .or. levelImage(k, j) .ne. 0_int16) cycle          
                if (NumberOfNeighbours22(levelimage, k, j, source%number) .lt. 2) cycle

        
!                if ( image(k, j) .ge. HeightFunction20(int(k), int(j), nSigma, &
!                     backgroundimage, backgroundSigmaImage)) then ! WAS 2.0

                 if (setLevelImagePixel(source, image, backgroundimage, &
                     backgroundSigmaImage, levelimage, k, j, nsigma)) then

        
                    source%imin = min(source%imin, k )
                    source%imax = max(source%imax, k)
                    source%jmin = min(source%jmin, j)
                    source%jmax = max(source%jmax, j)
                    levelImage(k, j) = source%number
                
                end if
  
            end do           
        end do

    end do
  
end subroutine RefineExtendedSourceMap2

!**************************************************************************************
!
!
!
!*********************************************************
logical Function setLevelImagePixel(source, image, &
    backgroundimage, backgroundSigmaImage, levelimage, i, j, nSigma)
    type(sourceParameters), intent(in):: source

 
    real(single), pointer :: image(:, :)
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)

    integer(int16), pointer :: levelImage(:, :)
    real(single), intent(in) :: nSigma
 
    integer(int16), intent(in) :: i, j
    real(single) :: xpos, ypos, radius, theta, dx, dy, r, dr
    real(single) :: cosTheta, sinTheta
    integer, parameter :: maxl = 500
    real(single) :: xl, xu, yl, yu
    integer :: k
    real(single) :: fb, fs, f1, f2, f3, f4, maxRadius
    real(single) :: constant, fmin
   

    if (abs(image(i, j) - backgroundimage(i, j)) .lt. 2.0) then
        setLevelImagePixel = .false.
            
        return
    end if
   

    if ( image(i, j) .lt. HeightFunction20(int(i), int(j), &
            nSigma, backgroundimage, backgroundSigmaImage)) then ! WAS 2.0
        setLevelImagePixel = .false.

        return
    end if

    if (NumberOfNeighbours22(levelimage, i, j, source%number) .lt. 2) then ! was 2
        setLevelImagePixel = .false.
            
        return
    end if

    xpos = real(i + 1)  ! x-pixel coord
    ypos = real(j + 1)  ! y-pixel coord
    dx = xpos - source%xpos 
    dy = ypos - source%ypos

    radius = sqrt(dx * dx + dy * dy)     ! Distance of current pixel from source origin

    if (radius .lt. real(4 / binx)) then
        setLevelImagePixel = .true.     
        return
    end if

    maxRadius = radius + real(6 / binx)  ! Upper radius limit


    theta = atan2(dy, dx)
    cosTheta = cos(theta)
    sinTheta = sin(theta)


    dr = 0.5 / real(binx)
    xl = 0.0
    xu = real(size(image, 1))
    yl = 0.0
    yu = real(size(image, 2))


    ! Interpolate at the pixel position and either side of it
   
    constant = 1.0
    fmin = 100000.0 

    do k = 4, 100
        r = real(k) * dr
        if(r .gt. radius + dr) then
            setLevelImagePixel = .true.
            return
        end if

        xpos = source%xpos + (r - dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit


        ypos = source%ypos + (r - dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f1 = Interpolate(image, xPos, yPos)
        fmin = min(fmin, f1)

        if (f1 .gt. fmin + sqrt(f1 + fmin)) then

            setLevelImagePixel = .false.
                      
            return
   
        end if

        fb = Interpolate( backgroundimage, xPos, yPos )
        fs = Interpolate( backgroundSigmaImage, xPos, yPos )

        !******************************
        ! Remove following for test
        !*****************************
        if (f1 .lt. max(5.0, fb + 2.0 * fs)) then        
        
            setLevelImagePixel = .false.
                      
            return
        end if
       

        xpos = source%xpos + r * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + r * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f2 = Interpolate(image, xPos, yPos)
      

        if (f2 .gt. f1 + constant * sqrt(f1 + f2)) then   ! Test on Friday
         
            setLevelImagePixel = .false.
                      
            return
        end if


        xpos = source%xpos + (r + dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + (r + dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
      
        f3 = Interpolate(image, xPos, yPos)
      
        if (f3 .gt. f1 + constant * sqrt(f1 + f3)) then
         
            setLevelImagePixel = .false.
                      
            return
        end if

        xpos = source%xpos + (r + 2.0 * dr) * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = source%ypos + (r + 2.0 * dr) * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
     

        f4 = Interpolate(image, xPos, yPos)
          
        if (f4 .gt. f1 + constant * sqrt(f1 + f4)) then
         
            setLevelImagePixel = .false.
                      
            return
        end if
 
    end do


    setLevelImagePixel = .true.       
end Function setLevelImagePixel



!***************************************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!***************************************************************************
real(single) Function HeightFunction20(i, j, nSigma, &
                      backgroundimage, backgroundsigmaimage)
    integer, intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage, backgroundSigmaImage


    HeightFunction20 = max(7.0, backgroundImage(i, j) + nSigma * backgroundSigmaimage(i, j)) ! Friday was 7.0


end Function  HeightFunction20


!***********************************************************
!
!***********************************************************
real(single) Function Interpolate(imagePtr, x, y)
    real(single), pointer :: imagePtr(:, :)
    real(single), intent(in) :: x, y

    real(single) :: xl, xu, yl, yu
    real(single) :: u, t
    integer j0, k0, j1, k1
    integer nx, ny

    nx = size(imagePtr, 1)
    ny = size(imagePtr, 2)

    xl = 0.0
    yl = 0.0
    xu = real(nx)
    yu = real(ny)

    if (x .gt. xl .and. x .lt. xu .and. & 
        y .gt. yl .and. y .lt. yu) then

        

        j0 = max(0, int(x - 0.5))
        j0 = min(nx - 1, j0)
      
        k0 = max(0, int(y - 0.5))
        k0 = min(ny - 1, k0)

        
        j1 = j0 + 1
        k1 = k0 + 1

        if (j1 .ge. nx) then
            t = 0.0
            j1 = nx - 1
        else        
            t = max(0.0, x - real(j0) - 0.5)
        end if

        if (k1 .ge. ny) then
            k1 = ny - 1
            u = 0.0
        else  
            u = max(0.0, y - real(k0) - 0.5)
        end if
      
        Interpolate = &
              (1.0 - t) * (1.0 - u) * imagePtr(j0, k0) + &
               t * (1.0 - u) * imagePtr(j1, k0) + &
               t * u * imagePtr(j1, k1) + &
               (1.0  - t) * u * imagePtr(j0, k1)

        Interpolate = max(0.0, interpolate)

        

    else
        Interpolate = 0.0
    end if


end function Interpolate

!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours22(levelimagePtr, i, j, number) result(count)
    integer(int16), pointer :: levelimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
    integer(int32), intent(in) :: number 
  
    integer(INT16) :: nx, ny

  
    integer(int8)  :: count
    integer(int16) :: il, iu, jl, ju, i1, j1
    ! Local variables
 
    nx = ubound(levelimagePtr, 1)
    ny = ubound(levelimagePtr, 2)

  
    il = max(0_int16, i - 1_int16)
    jl = max(0_int16, j - 1_int16)

  
    iu = min(nx, i + 1_int16)
    ju = min(ny, j + 1_int16)


 

    count = 0
    do i1 = il, iu
        do j1 = jl, ju
            if(i1 .eq. i .and. j1 .eq. j) cycle
            if (levelImagePtr(i1, j1) .eq. number) count = count + 1
        end do
    end do 
  


end function NumberOfNeighbours22


!**********************************************************************
! RefineSourceMap4 - For bright sources
!
!
!**********************************************************************
subroutine RefineSourceMap4(source, backgroundImage, &
             maskImage, emaskimage)
  
  
    type(sourceParameters), intent(inout) :: source 
 
    real(single), pointer, dimension(:, :) :: backgroundImage
   
    logical(boolean), pointer, dimension(:, :) :: maskImage, emaskimage
   
    integer(int16) :: nx, ny
    integer :: maxPos(2)
    integer(int16) :: il, iu, jl, ju, dx 
    integer(int16) :: i0, j0, l, m, n, dl

    real(single), parameter :: nstd = 3.0
    
    nx = ubound(backgroundimage, 1)
    ny = ubound(backgroundimage, 2)


    dx = 12 / binx

    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)

    maxPos = maxloc(backgroundimage(il : iu, jl : ju))

    i0 = maxPos(1) - 1 + il
    j0 = maxPos(2) - 1 + jl
    dl = 12 / binx
    dx = 100 / binx



    il = max(source%imin - dx, int(0, int16))
    iu = min(source%imax + dx, nx)
    jl = max(source%jmin - dx, int(0, int16))
    ju = min(source%jmax + dx, ny)


    do l = 1, dx
        il = max(int(0, int16), i0 - l)
        iu = min(nx, i0 + l)
        jl = max(int(0, int16), j0 - l)
        ju = min(ny, j0 + l)


        do n = jl, ju
            do m = il, iu

                if (emaskImage(m, n)) cycle
                if (l .le. dl) then
               
                    maskImage(m, n) = .true.    
                    emaskImage(m, n) = .true.    
                         
                else if (setMaskPixel(i0, j0, backgroundimage, m, n)) then
                                
                    maskImage(m, n) = .true.    
                    emaskImage(m, n) = .true.                 
                end if  
            end do            
        end do

    end do
  

!   call saveimage(levelimage, nx + 1, ny + 1, "levelimage.fits", "testimage")

!call fatal("stopped line 8461 displaying levelimage.fits")


  
end subroutine RefineSourceMap4

!**************************************************************************************
!
!
!
!*************************************************************************************
logical Function setMaskPixel(i0, j0, backgroundimage, i, j)
    integer(int16), intent(in) :: i0, j0
    real(single), pointer :: backgroundImage(:, :)
    integer(int16), intent(in) :: i, j
    real(single) :: xpos, ypos, radius, maxradius, radius1, theta, dx, dy, r, dr
    real(single) :: cosTheta, sinTheta
    real(single) :: xl, xu, yl, yu
    integer :: k
    real(single) :: f1, f2
    real(single) :: xCentre, yCentre
    real(single), parameter :: df = 0.003
!integer, parameter :: ipix = 6, jpix=-434

    xCentre = real(i0) + 0.5  ! Centre x pixel coord
    yCentre = real(j0) + 0.5  ! Centre y pixel coord

    xpos = real(i) + 0.5 ! x-pixel coord
    ypos = real(j) + 0.5 ! y-pixel coord
  
    dx = xpos - xCentre
    dy = ypos - yCentre


    radius1 = sqrt(dx * dx + dy * dy)     ! Distance of current pixel from source origin

    if (radius1 .lt. real(4 / binx)) then
        setMaskPixel = .true.     
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1598" 

    theta = atan2(dy, dx)
    cosTheta = cos(theta)
    sinTheta = sin(theta)


    dr = 1.0 / real(binx)
    maxRadius = radius1 + real(6 / binx)  ! Upper radius limit


    xl = 0.0
    xu = real(size(backgroundimage, 1))
    yl = 0.0
    yu = real(size(backgroundimage, 2))

 
    ! Interpolate at the pixel position and either side of it
 
    radius = 3.0 * dr
     
    xpos = xCentre + radius * cosTheta
    if (xpos .lt. xl .or. xpos .gt. xu) then
        setMaskPixel = .true.
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1628 i0=", i0, " j0=", j0 

    ypos = yCentre + radius * sinTheta

    if (ypos .lt. yl .or. ypos .gt. yu) then 
        setMaskPixel = .true.
        return
    end if

!       if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1637" 


    f1 = Interpolate(backgroundimage, xPos, yPos)
  
    if (f1 .lt. 0.1) then
        setMaskPixel = .false.
        return
    end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1647" 


    do k = 4, 100
       
        r = real(k) * dr
        if(r .gt. maxradius) then
            setMaskPixel = .true.
            return
        end if
!        if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1656 r=", r 


        xpos = xCentre + r * cosTheta
        if (xpos .lt. xl .or. xpos .gt. xu) exit
        ypos = yCentre + r * sinTheta
        if (ypos .lt. yl .or. ypos .gt. yu) exit
       
        f2 = Interpolate(backgroundimage, xPos, yPos)

!      if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1666 f1=", f1, " f2=", f2 

        if (f2 .lt. 0.1) then
            setMaskPixel = .false.
                      
            return
        
        end if

!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1673 f1=", f1, " f2=", f2 

        
        if (  abs(f1 - f2) / f1  .lt. df ) then

            setMaskPixel = .false.
                      
            return
    
        
        end if
!if (i .eq. ipix .and. j .eq. jpix) write(*, *) "Line 1684 f1=", f1, " f2=", f2 

     
        f1 = f2
    end do

    setMaskPixel = .true.
       
end Function setMaskPixel

end module Omsourcemap_mod


