!*************************************************************
! Module sourcedetection_mod
! Created 15th April 2002- C.Brindle
! Contains a subroutine for detecting "point" sources and one 
! for detecting any other kind of source
!****************************************************************
module psffunctions_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmConstants_mod      ! For binx, biny, etc
    use omsmoothimage_mod           ! For polynomial smoothing
    use omnumerics_mod 
    use leastsquares_mod
    use ominterpolate_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod
    use gridinterpolation_mod   
    use ompsf_mod



type PsfParameters
    real(single) :: psfRatio(3)
    real(single), pointer :: psfImage(:, :) => null()
    real(single) :: psfSemiA = -1.0, psfsemiB = -1.0
    real(single) :: sigmax = -1.0, sigmay = -1.0
end type PsfParameters

type(PsfParameters), save :: psfParams 


interface PSFPhotometry
    module procedure PSFPhotometry1, PSFPhotometry2
end interface PSFPhotometry

contains

!*****************************************************
! Computes the semia and semib  of the psf using
! moments 
!
!*****************************************************
subroutine GetPSFWidths(semiA, semiB)

    real(single), intent(out) :: semiA, semiB


    integer(int32) :: psfNx, psfNy
    type(sourceParameters) :: source
   
    real(single), pointer :: psfImage1(:, :) => null()


    if (associated(psfParams%psfImage)) return

    psfnx = 12
    psfny = 12


    ! Get the point-spread function image
  



    call CreatePSFImage21(psfNx, psfNy, psfParams%psfImage, 1024.5_double, 1024.5_double, &
        ratio=0.5_double, bin=binX, res=1) 

    
    psfImage1 => allocate2dSingle(0, psfnx - 1, 0, psfny - 1, "psfimage1")


    source%xPos = real(psfNx) / 2.0
    source%yPos = real(psfNy) / 2.0
       
    psfImage1 = psfParams%psfImage * 100.0

    ! Compute the semi-major and minor axes of it


    call ComputePointSourceParameters(int(-1, int16), int(psfNx - 1, int16), int(psfNy - 1, int16), psfImage1, &
                         int(psfNx / 2 - 1, int16), int(psfNy / 2 - 1, int16), source, int(4 / binx, int8))


!    write(*, '( a, f8.2, a, f6.2, a, f8.2, a, f6.2)') "Computed PSF semia=", source%semia, "+/-", source%semiaerr,  &
!       " b=", source%semib, " +/-", source%semibErr

    semiA = source%semiA
    semiB = source%semiB

    deallocate(psfimage1)    
    nullify(psfimage1)


    !*****************************************************************************
    ! Compute the FWHM 
    !
    !*****************************************************************************

!    fwhm 


!    write(*, *) "Computed a=", computedMoments%semia, "+/-", computedMoments%semiaErr, &
!         " b=", computedMoments%semib, "+/-", computedMoments%semibErr


!call FitPSfFunction(source, psfImage, binX, 12 / binx)
!call FitPSfFunction(source, psfImage, binX, 8 / binx)


!call Fit2dGaussian2(source, psfImage, binX)
  
      

end subroutine GetPSFWidths

!******************************************************************
!
!
!*****************************************************************
subroutine ClassifySource(source, image, backgroundImage)
    
    type(sourceParameters), intent(inout) :: source
    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16) :: i0, j0 
    real(single), parameter :: nstd = 3.0 
    integer(int16) :: il, iu, jl, ju
    source%semiaErr = max(0.1,  source%semiaErr) ! Was 0.02
    source%semibErr = max(0.1,  source%semibErr)

         
    ! Check classification

    il = lbound(image, 1)
    jl = lbound(image, 2)
    iu = ubound(image, 1)
    ju = ubound(image, 2)


    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)
    i0 = max(il, i0)
    j0 = max(jl, j0)
    i0 = min(iu, i0)
    j0 = min(ju, j0)
!write(*, *) "Classify source x=", source%xpos, " y=", source%ypos, " semia=", source%semia, " error=", source%semiaerr, &
!" psfsemia=", psfparams%psfsemia    

    if ((image(i0, j0) - backgroundImage(i0, j0)) .lt. 2000.0 .and. & 
            (source%semia - psfparams%psfsemia) .gt. (nstd * source%semiaErr)) then


        source%extensionFlag = 1
        
              
    else
        source%extensionFlag = 0
    end if
!write(*, *) "extension=", source%extensionFlag

end subroutine ClassifySource


!*************************************************************************
!
!
!
!************************************************************************

subroutine PSFPhotometry3(image, psfImage, backgroundimage, source, dx1)
    real(single), dimension(:, :), pointer :: image, psfImage, backgroundimage

    type(SourceParameters), intent(inout) :: source
    integer, intent(in) :: dx1

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    real(single), pointer, dimension(:) :: xdata => null(), ydata => null(),  dydata => null(), weight => null()
    real(single) :: psfxCentre, psfYcentre
    real(single) :: xpos, ypos, dx, dy
    real(single) :: gradient, gradientError, constant, constantError

!call CentroidSource(source, image, backgroundimage)

!return



    psfnx = size(psfimage, 1)
    psfny = size(psfimage, 2)

    nx = size(image, 1)
    ny = size(image, 2)


    i0 = psfnx / 2
    j0 = psfny / 2
    imin =  max(0, i0 - dx1)
    imax =  min(i0 + dx1, psfnx - 1)
    jmin =  max(0, j0 - dx1)
    jmax =  min(j0 + dx1, psfny - 1)

    n = (imax - imin + 1) * (jmax - jmin + 1)

    xdata => allocateSingle(1, n, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    dydata => allocateSingle(1, n, "dydata")
    weight => allocateSingle(1, n, "weight")


    psfXcentre = real(psfnx) / 2.0
    psfYcentre = real(psfny) / 2.0

!write(*, *) "source pos x=", source%xpos, " y=", source%ypos

    n = 0



    do i = imin, imax
        dx = real(i0 - i)
        xpos = source%xpos - dx
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(nx - 1)) cycle 
        do j = jmin, jmax
            dy = real(j0 - j) 
            ypos = source%ypos - dy


            if (ypos .lt. 0.0 .or. ypos .gt. real(ny - 1)) cycle 
            n = n + 1
            xdata(n) = psfImage(i, j)
            ydata(n) =  bilinearInterpolate(image, xPos, yPos) -  bilinearInterpolate(backgroundimage, xPos, yPos)
            dydata(n) = sqrt(ydata(n))
!            write(*, *) "i=", i, " j=", j, " dx=", dx, " dy=", dy, " x=", x(n), " y=", y(n), "xpos=", xpos, " ypos=", ypos


           weight(n) = 1.0 / (1.0 + sqrt(dx * dx + dy * dy))
           if (ydata(n) .gt. 0.0) then
               weight(n) = 1.0 / ydata(n)
           else
               weight(n) = 1.0
           end if
        end do
    end do

 !   do i = 1, n
 !       write(*, *) i, " x=", x(i), " y=", y(i)
 !   end do



call LeastSquaresFit1(n, xdata, ydata, weight, gradient, gradientError)

    call LeastSquaresFit(xdata, ydata, dydata, gradient, gradientError, constant, constantError)

    deallocate(xdata, ydata, dydata, weight)
    nullify(xdata, ydata, dydata, weight)

!do i = 10, 15
!    xpos = real(i) + 0.5
!    ypos = xpos
!    write(*, *) "xpos==", xpos, " ypos=", ypos, " true=", image(i, i),  " interpolates=",Interpolate(image, xPos, yPos) 
!end do

end subroutine PSFPhotometry3


!*************************************************************************
!
!
!
!************************************************************************

subroutine PSFPhotometry1(image, psfImage, levelImage, backgroundimage, source)
    real(single), dimension(:, :), pointer :: image, psfImage, backgroundimage
    integer(int16), dimension(:, :), pointer :: levelImage
    integer :: i

    type(SourceParameters), pointer :: source(:)
    
    do i = 1, size(source)
        if (source(i)%extensionFlag .eq. 1) cycle

        call PSFPhotometry2(image, psfImage, levelImage, backgroundimage, source(i))

    end do
end subroutine PSFPhotometry1

!*************************************************************************
!
!
!
!************************************************************************

subroutine PSFPhotometry2(image, psfImage, levelImage, backgroundimage, source)
    real(single), dimension(:, :), pointer :: image, psfImage, backgroundimage
    integer(int16), dimension(:, :), pointer :: levelImage


    type(SourceParameters), intent(inout) :: source
    

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    real(single), pointer, dimension(:) :: xdata => null(), ydata => null(),  dydata => null(), weight => null()
    real(single) :: psfxCentre, psfYcentre
    real(single) :: xpos, ypos, dx, dy
    real(single) :: gradient, gradientError, constant, constantError
  
!call CentroidSource(source, image, backgroundimage)

!return



    psfnx = size(psfimage, 1) 
    psfny = size(psfimage, 2)  

    nx = size(image, 1)
    ny = size(image, 2)


    i0 = psfnx / 2
    j0 = psfny / 2
    imin =  source%imin
    imax =  source%imax
    jmin =  source%jmin
    jmax =  source%jmax

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocateSingle(1, n, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    dydata => allocateSingle(1, n, "dydata")
    weight => allocateSingle(1, n, "weight")


    psfXcentre = real(psfnx) / 2.0 ! The x centre of the psf image
    psfYcentre = real(psfny) / 2.0 ! The y centre of the psf image

!write(*, *) "source pos x=", source%xpos, " y=", source%ypos

    n = 0

    write(*, *) "Sum counts in the psfimage=", sum(psfimage)

    do i = imin, imax
        dx = source%xpos - real(i + 1)
        xpos = psfXcentre + dx
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
        do j = jmin, jmax
            if (levelImage(i, j) .ne. source%number) cycle

 
            dy = source%ypos - real(j + 1)
     
            ypos = psfYcentre + dy
  


            if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
            n = n + 1
           

            xdata(n) =  bilinearInterpolate(psfimage, xPos, yPos)
            if (xData(n) .le. 0.0) then
                n = n - 1
                cycle
            end if
            ydata(n) =  image(i, j) - backgroundimage(i, j)


            dydata(n) = sqrt(ydata(n))


!            write(*, *) "i=", i, " j=", j, " dx=", dx, " dy=", dy, " x=", x(n), " y=", y(n), "xpos=", xpos, " ypos=", ypos


           weight(n) = 1.0 / (1.0 + sqrt(dx * dx + dy * dy))
           if (ydata(n) .gt. 0.0) then
               weight(n) = 1.0 / ydata(n)
           else
               weight(n) = 1.0
           end if
        end do
    end do

 !   do i = 1, n
 !       write(*, *) i, " x=", x(i), " y=", y(i)
 !   end do


  
   call LeastSquaresFit1(n, xdata, ydata, weight, gradient, gradientError)
  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

 




    call LeastSquaresFit(xdata, ydata, dydata, gradient, gradientError, constant, constantError)

  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

    deallocate(xdata, ydata, dydata, weight)
    nullify(xdata, ydata, dydata, weight)

!do i = 10, 15
!    xpos = real(i) + 0.5
!    ypos = xpos
!    write(*, *) "xpos==", xpos, " ypos=", ypos, " true=", image(i, i),  " interpolates=",Interpolate(image, xPos, yPos) 
!end do

end subroutine PSFPhotometry2


end module psffunctions_mod


