!*************************************************************
!
!*************************************************************
MODULE sourceanalysis_mod

    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSkyBackground_mod  ! For computing sky background mode
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use omparameters_mod     ! For computing source moments
    use OmSmoothImage_mod    ! For smoothing image
    use OmConstants_mod      ! For binx, biny, etc
    use ompsf_mod
    use omdetectdetection_mod
    use cubicshepard_mod
IMPLICIT NONE



!***************************************************************
! For extended-source detection
!**************************************************************
integer(int16), pointer, private, dimension(:, :) :: tempelevelImage => null()
real(single), pointer, private, dimension(:, :) :: tempBackgroundImage => null()


integer(int8), private :: minCounts    

type(SourceLinkedList), private, save :: tempESourceList  ! The linked-list point-source list

integer(int16), pointer, private, dimension(:, :) :: templevelImage => null()
integer(int16), pointer, private, dimension(:, :) :: maxtemplevelImage => null()
integer(int32), private :: minNopixels, maxNoPixels

integer(int16), private :: kmax, maxnsources


CONTAINS



!***********************************************************
! Checks extended sources and tries to split each one
! into 2 or more sources 
!
!***********************************************************
subroutine decomposeExtendedSources(image,           &
                                    plevelImage,     &
                                    elevelImage,     & 
                                    eSourceList,     &
                                    nESources,       &
                                    isValidSource)

    real(single), pointer, dimension(:, :) :: image
    integer(int16), pointer, dimension(:, :) :: plevelimage, elevelImage
    type(SourceLinkedList), intent(inout):: eSourceList  ! The linked-list extended-source list
    integer(int16), intent(inout) :: neSources   
    type(sourceParameters), pointer :: esource(:) => null()      ! The extended sources
    integer :: i  

    interface

        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:)
            logical(bool) :: status

        end function

    end interface




    if (nESources .eq. 0) return
    write(errstr,'(a, i5)') "Reanalysing extended sources- no=", nESources
    call message(AppMsg, Verbose)
 
 
    nx = size(image, 1)
    ny = size(image, 2)
    
    ! Load the extended sources
    esource => SourceDataArray(esourceList, neSources)
    call DeleteSourceList(eSourceList)

  

    tempeLevelImage => allocate2dInt16(0, nx - 1, 0, ny - 1, "tempelevelimage")
  
    tempBackgroundImage => allocate2dSingle(0, nx - 1, 0, ny - 1, "tempBackgroundimage")
   
    kmax = 20
   

    tempLevelImage => allocate2dInt16(0, nx - 1, 0, ny - 1, "templevelimage")   
    maxtempLevelImage => allocate2dInt16(0, nx - 1, 0, ny - 1, "maxtemplevelimage")

  

    !******************************************************************
    ! Loop through all the extended sources and see if they can be 
    ! decomposed into two or more sources
    !
    !
    !*****************************************************************
    neSources = 0    
    title = ""

    do i = 1, size(esource)  ! Loop through all the extended sources 
!write(*, *) i, " number=", esource(i)%number, " x=", esource(i)%xpos, " y=", esource(i)%ypos
!if (abs(esource(i)%xpos - 25.0) .gt. 8.0 .or. abs(esource(i)%ypos - 165.0) .gt. 8.0) then


        if (esource(i)%noSourcePixels .lt. 40 / binX .or. &
            esource(i)%noSourcePixels .gt. 50000 / binX .or. &
            esource(i)%extensionFlag .eq. 0) then


            ! Don't try decomposing this source
            neSources = neSources + 1


            where(eLevelImage(esource(i)%imin : esource(i)%imax, esource(i)%jmin : esource(i)%jmax) &
                   .eq. esource(i)%number) &
                 tempeLevelImage(esource(i)%imin : esource(i)%imax, esource(i)%jmin : esource(i)%jmax) = -neSources 

            eSource(i)%number = -neSources ! The new source number         

            call AddSourceToList(esourceList, eSource(i))
         
        else

!write(*, *) "Line 294 trying to deconvolve source ", i, " source no=", esource(i)%number
            !************************************************************
            ! Try to decompose this source into 2 or more sources
            !***********************************************************
            call decomposeExtendedSource(esource(i),  &
                     image,                           &
                     plevelimage,                     &
                     elevelImage,                     &
                     esourceList,                     & 
                     neSources,                       &           
                     isValidSource)                                                                                 

        end if

    
    end do

!  write(*, *) "line 325 minval elevelimage =", minval(elevelimage)
! write(*, *) "line 325 minval tempelevelimage =", minval(tempelevelimage)

    write(errstr,'(a, i5)') "New number of extended sources= ", neSources
    call message(AppMsg, Verbose)

    deallocate(esource)
    nullify(esource)
            
  
    elevelImage = tempElevelImage


!  call saveimage(tempelevelimage, nx, ny, "templevelimage.fits", "testimage")
!call fatal("stopped at line 403 displaying templevelimage.fits")

    deallocate(tempBackgroundImage, tempElevelImage)
    nullify(tempBackgroundImage, tempElevelImage)
              

            
    deallocate(tempLevelImage, maxtemplevelImage)
    nullify(tempLevelImage, maxtemplevelImage)
 
 


end subroutine DecomposeExtendedSources


!***********************************************************
! Checks extended sources and tries to split each one
! into 2 or more sources 
!
!***********************************************************
subroutine decomposeExtendedSource(esource,             &
                                   image,               &
                                   plevelimage,         &
                                   elevelImage,         &
                                   eSourceList,         &
                                   nESources,           &
                                   isValidSource)

    type(sourceParameters), intent(inout) :: esource      ! The extended sources
    real(single), pointer :: image(:, :)
    integer(int16), pointer, dimension(:, :) :: plevelimage, elevelImage
    type(SourceLinkedList), intent(inout):: eSourceList   ! The linked-list extended-source list
    integer(int16), intent(inout) :: neSources   
    integer(int16) :: i, j, jmax
    real(single) :: height, minImage, maxImage, heightRange
    real(double) :: sum
    integer(int16) ::nSources    
    integer(int16) :: sourceNumber
    type(sourceParameters), pointer :: tempesource(:) => null()  
    real(single) :: minimumCounts 
    integer(int16) :: nx, ny
    interface

        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:)
            logical(bool) :: status

        end function

    end interface





    minNoPixels = 20 / binx
    minCounts = 2

! write(*, *) "Line 347 source no=", esource%Number
!write(*, *) "xpos=", esource%xpos, " y=", esource%ypos
!write(*, *) "imin=", esource%imin, " imax=", esource%imax, &
!" jmin=", esource%jmin, " jmax=", esource%jmax           

    nx = size(image, 1)
    ny = size(image, 2)


    if (esource%noSourcePixels .lt. 2 * minNoPixels .or. &
!        esource%noSourcePixels .gt. 10000 / binX .or. &
        esource%extensionFlag .eq. 0) then
 
        ! Don't try decomposing this source
        neSources = neSources + 1
        sourceNumber = -nesources

        where(eLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
                   .eq. esource%number) &
             tempeLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = sourceNumber 

        eSource%number = sourceNumber ! The new source number


!write(*, *) "Line 360 adding source no ", esource%number, " to list"
!write(*, *) "xpos=", esource%xpos, " y=", esource%ypos
!write(*, *) "imin=", esource%imin, " imax=", esource%imax, " jmin=", esource%jmin, " jmax=", esource%jmax           
        call AddSourceToList(esourceList, eSource)
        return ! Finished        
    end if 
        
    maxNoPixels = esource%noSourcePixels - minNoPixels  

    minImage = 1000000.0
    maximage = 0.0
 
    sum = 0.0_double
 
    do j = esource%jmin, esource%jmax
        do i = esource%imin, esource%imax
            if (elevelImage(i, j) .ne. esource%number) cycle
            if (image(i, j) .ge. 0.0) sum = sum + dble(image(i, j))
            if (image(i, j) .lt. minImage) then
                minImage = image(i, j)
            else if (image(i, j) .gt. maxImage) then
                maximage = image(i, j)
            end if

        end do
    end do 

    minimumCounts = sum / 10.0_double ! The minimum number of counts allowed in a split source
    call setMinimumCounts( minimumCounts ) 

    ! Set the background image to the minimum value within the extended region
    tempbackgroundImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = minimage


    heightRange = (maxImage - minImage) * 0.75         
    
    heightRange = log10(heightrange) / real(kmax)
   
 
    where(elevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
            .eq. esource%number)
        tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 0
    elsewhere
        templevelimage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 1
    end where
        


! call saveimage(templevelimage, nx, ny, "levelimage.fits", "testimage")
!call fatal("stopped at line 419 displaying levelimage.fits")



    !*********************************************************************************
    ! Find the maximum number of sources that the extended source can be split into
    !*********************************************************************************
    maxNSources = 0
    jMax = 1
 
    maxtempLevelImage = 0


    do j = 1, kmax


        height = 10.0 ** (real(j) * heightRange) 
        if ((minImage + height) .gt. maxImage) exit

!write(*, *) "Iteration ", j, " height=", height 
        ! Run the detection algorithim on the extended source region
        
        nSources = 0
          
        call DetectExtendedSources(nx, ny, esource%imin, esource%imax,             &
                   esource%jmin, esource%jmax,                                     &
                   image, image, tempbackgroundImage,                              &
                   maxImage, plevelImage, tempLevelImage,                          &   
                   tempeSourceList, 1_int16, nSources, minNoPixels, maxNoPixels,   &      
                   minCounts,  height, heightFunction2, IsValidSource,             &
                   int(1, int8), .false.)            

        if (nSources .gt. maxNSources) then
            maxNSources = nSources
            jmax = j
            maxtempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = &
                 tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax)
        
            if (associated(tempESource)) deallocate(tempEsource)
            ! Load the extended sources
            tempEsource => SourceDataArray(TempesourceList, nSources)

            

        end if


        where (tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) .ne. 1) &
                   tempLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = 0

            
        call DeleteSourceList(tempeSourceList)
   
    end do 

    if (maxNSources .lt. 2) then ! Not split into 2 or more sources 
        neSources = neSources + 1
        sourceNumber = -neSources  ! The next source number
        where(eLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) &
               .eq. esource%number) &
                tempeLevelImage(esource%imin : esource%imax, esource%jmin : esource%jmax) = sourceNumber
        esource%number = sourceNumber


        call AddSourceToList(esourceList, eSource)
        return
    end if
    
       
    write(errstr,'(a, i5, a, f9.2, a, f9.2, a, i2, a)') "Source ", esource%number,   &
           " (", esource%xpos, ", ", esource%ypos,                                   &
           ") has been split into ", maxNSources, " sources"
    call message(AppMsg, Verbose)
                            
         

!    write(errstr,'(a)') "New sources"
!    call message(AppMsg, Verbose)
      
    ! Add to the extended-source list   
    do j = 1, maxNSources            
        neSources = neSources + 1
           
        sourceNumber = -neSources
     

        where(maxtempLevelImage(tempeSource(j)%imin : tempeSource(j)%imax, tempESource(j)%jmin : tempESource(j)%jmax)&
              .eq. tempesource(j)%number) &
               tempeLevelImage(tempeSource(j)%imin : tempeSource(j)%imax, tempESource(j)%jmin : tempESource(j)%jmax) = sourceNumber 

        tempESource(j)%number = sourceNumber

!        write(errstr,'(a, i5, a, f9.2, a, f9.2)') "Source ", tempesource(j)%number, &
!           " (", tempesource(j)%xpos, ", ", tempeSource(j)%ypos    
 
!        call message(AppMsg, Verbose)
      
!write(*, *) "imin=", tempeSource(j)%imin, " imax=", tempesource(j)%imax, &
!            " jmin=",  tempeSource(j)%jmin, " jmax=", tempesource(j)%jmax

        call AddSourceToList(esourceList, tempeSource(j))

       
            !************************************************************
            ! Try to decompose this source into 2 or more sources
            !***********************************************************
!            call decomposeExtendedSource(tempEsource(j),      &
!                                         image,               &
!                                         tempelevelImage,     &
!                                         eSourceList,         &
!                                         nESources) 
 
    end do

!if (esource%number .eq. -8) then
!  call saveimage(maxtemplevelimage, nx, ny, "levelimage.fits", "testimage")
!call fatal("stopped at line 589 displaying levelimage.fits")

!end if
! call saveimage(tempelevelimage, nx, ny, "levelimage.fits", "testimage")
!call fatal("stopped at line 589 displaying levelimage.fits")


    deallocate(tempESource)
    nullify(tempESource)


end subroutine DecomposeExtendedSource



end module sourceanalysis_mod    








